#include <SPI.h>
#include <Wire.h>
#include "U8glib.h"

//https://rntlab.com/question/analogread-for-ntc-thermistor/
//https://learn.adafruit.com/thermistor/using-a-thermistor
//https://www.electronicwings.com/nodemcu/thermistor-interfacing-with-nodemcu
//NTC 100k thermistor 3950
//ESP8266 チップ本体はadc 1V max、モジュールは3.3V max
//https://randomnerdtutorials.com/esp8266-adc-reading-analog-values-with-nodemcu/

const double VCC = 5.0;             // NodeMCU on board 3.3v vcc
const double R2 = 4650; //100000;//97680; //100000;            // 10k ohm series resistor
const double adc_resolution = 1023; // 10-bit adc
// resistance at 25 degrees C
const double THERMISTORNOMINAL = 100000;//92800;
// temp. for nominal resistance (almost always 25 C)
const double  TEMPERATURENOMINAL = 25;
// The beta coefficient of the thermistor (usually 3000-4000)
const double  BCOEFFICIENT = 3950;

unsigned long last_tempcheck;
unsigned long current_millis;
unsigned long time_over_220 = 0;
char chbuff[64];
float tempavg[5];
int tempidx = 0;

#define BUTTON_PIN 8 //connect switch, and to gnd
#define RELAY_PIN 9

/*  #define SCREEN_WIDTH 128 // OLED display width, in pixels
    #define SCREEN_HEIGHT 64 // OLED display height, in pixels
    #define OLED_RESET     -1 // Reset pin # (or -1 if sharing Arduino reset pin)
    Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire, OLED_RESET);*/
//OLEDディスプレイ
//u8glib フォント https://github.com/olikraus/u8glib/wiki/fontsize
U8GLIB_SSD1306_128X64 u8g(U8G_I2C_OPT_NONE | U8G_I2C_OPT_DEV_0);   // I2C / TWI これにしないとフリーズする

char oledstrbuff[64];

const unsigned char bmpopen[] PROGMEM = {
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xB8, 0x3B, 0x00, 0x00, 0xA8, 0x08, 0x00, 0x00, 0xA8, 0x19, 0x00, 0x00, 0xA8, 0x08, 0x00, 0x00, 0xB8, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0x01, 0x80, 0xFF, 0xFF, 0x03, 0x80, 0xFF, 0xFF, 0x03, 0xE0, 0xFF, 0xFF, 0x0F, 0x10, 0x00, 0x80, 0x16, 0x10, 0x00, 0x80, 0x16, 0x10, 0x00, 0x80, 0x16, 0x10, 0x00, 0x80, 0x16, 0xF0, 0xFF, 0xFF, 0x1F, 0x08, 0x00, 0x00, 0x20, 0x08, 0x00, 0x00, 0x20, 0x08, 0x00, 0x00, 0x20, 0x08, 0x00, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
};
const unsigned char bmpclose[] PROGMEM = {
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x81, 0x00, 0x00, 0x81, 0x40, 0x00, 0x80, 0x40, 0x20, 0x00, 0x80, 0x40, 0x20, 0x00, 0x00, 0x01, 0x40, 0x00, 0x00, 0xF2, 0x92, 0x00, 0x00, 0x94, 0x16, 0x01, 0x00, 0x94, 0x1A, 0x01, 0x00, 0x92, 0x92, 0x00, 0x00, 0xF1, 0x52, 0x00, 0x80, 0x00, 0x00, 0x00, 0x80, 0x40, 0x20, 0x00, 0x00, 0x81, 0x40, 0x00, 0x00, 0x02, 0x81, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0x01, 0x80, 0xFF, 0xFF, 0x03, 0x80, 0xFF, 0xFF, 0x03, 0xE0, 0xFF, 0xFF, 0x0F, 0x10, 0x00, 0x80, 0x16, 0x10, 0x00, 0x80, 0x16, 0x10, 0x00, 0x80, 0x16, 0x10, 0x00, 0x80, 0x16, 0xF0, 0xFF, 0xFF, 0x1F, 0x08, 0x00, 0x00, 0x20, 0x08, 0x00, 0x00, 0x20, 0x08, 0x00, 0x00, 0x20, 0x08, 0x00, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
};


//https://www.ablic.com/jp/semicon/support/package/solder-temp-profile/
//温度管理
//加熱
//150度～180度　90秒
//加熱　220度
//ピーク　260度
//220度から再度220度まで30～40秒
//クールダウン

#define TYPE_ON	1	//通電
#define TYPE_OFF	2	//切断
#define TYPE_KEEP	3	//保温
typedef struct TEMP_DATA {
	int type;		//種別
	int on;		//オンにする温度、もしくは設定温度
	int off;		//オフにする温度
	unsigned long duration;	//0にすると指定温度到達まで継続
	char msg[32];
};

#define NUM_DATA 7
TEMP_DATA tdata[NUM_DATA] = {
	{TYPE_ON,		130,	0,	0, "Power up"},	//150度までオン 急激に上がるので手前で止めておく
	{TYPE_KEEP,	130,	135, 30, "Preheat 1 130\xb0"}, //120秒間少しずつ温度を上げる
	{TYPE_KEEP,	140,	145, 30, "Preheat 2 140\xb0"},
	{TYPE_KEEP,	145,	155, 30, "Preheat 3 150\xb0"},
	{TYPE_KEEP,	160,	165, 30, "Preheat 4 160\xb0"},
	{TYPE_ON,		200,	0,	0, "Heating"},
	{TYPE_OFF,	40,	0,	0, "Cooldown"},	//オフ
};
bool controlling = false;
int control_index;
unsigned long control_start = 0;
bool relay_state = false;

int temp_hist[128];
int temp_index = 0;

void oled_update(int temp) {
	u8g.setColorIndex(1);
	u8g.firstPage();
	if (time_over_220 > 0)
		sprintf(chbuff, "Temp:%d\xb0 [%ds]", (int)temp, (int)((millis() - time_over_220)/1000));
	else
		sprintf(chbuff, "Temp:%d\xb0", (int)temp);
	if (controlling && control_index < NUM_DATA) {
		if (tdata[control_index].duration > 0) {
			sprintf(chbuff, "Temp:%d\xb0  %ds", (int)temp, tdata[control_index].duration - (current_millis - control_start) / 1000);
		}
	}
	do {
		u8g.setFont(u8g_font_6x12);
		u8g.drawStr(0, 10, "Reflow Control");
		u8g.drawStr(0, 20, chbuff);
		if (controlling && control_index < NUM_DATA) {
			u8g.drawStr(0, 30, tdata[control_index].msg);
		}
		if (relay_state)
			u8g.drawXBMP( 97, 0, 32, 32, bmpclose);
		else
			u8g.drawXBMP( 97, 0, 32, 32, bmpopen);
		for (int i = 0; i < 128; i++) {
			if (temp_hist[(temp_index + i) % 128] > 0)
				u8g.drawPixel(i, max(32, 63 - (temp_hist[(temp_index + i) % 128] / 10)));
		}

	} while ( u8g.nextPage() );
}

float gettemp() {
	double Vout, Rth, temperature, adc_value, steinhart;
	adc_value = analogRead(A0);
	Serial.println(adc_value);
	Vout = (adc_value * VCC) / adc_resolution;
	Rth = (R2 / (VCC - Vout)) * Vout;

	steinhart = Rth / THERMISTORNOMINAL;     // (R/Ro)
	steinhart = log(steinhart);                  // ln(R/Ro)
	steinhart /= BCOEFFICIENT;                   // 1/B * ln(R/Ro)
	steinhart += 1.0 / (TEMPERATURENOMINAL + 273.15); // + (1/To)
	steinhart = 1.0 / steinhart;                 // Invert
	temperature = steinhart - 273.15;                         // convert absolute temp to C

	tempavg[tempidx] = temperature;
	tempidx = (tempidx + 1) % 5;
	temperature = (tempavg[0] + tempavg[1] + tempavg[2] + tempavg[3] + tempavg[4]) / 5;
	Serial.print("  ");
	Serial.println(temperature);

	return temperature;
}

bool buttonstate = false;
unsigned long button_time = 0;
bool buttonpress() {
	if (digitalRead(BUTTON_PIN) == LOW) {
		if (!buttonstate) {
			Serial.println("btn down");
			delay(50);
			if (digitalRead(BUTTON_PIN) == LOW) { //再チェック
				Serial.println("btn down confirm");
				buttonstate = true;
				button_time = millis();
			}
		}
	} else if (buttonstate) {
		Serial.println("btn up");
		delay(50);
		buttonstate = false;
		if (!controlling) {
			if (current_millis - button_time >= 700) //少し長押しして反応
				return true;
		} else {
			return true;
		}
	}
	return false;
}

unsigned long last_tempdraw = 0;
unsigned long last_temptext = 0;
void draw_tempgraph(float temp) {
	//32 dot - 64 dot
	//32dot 320deg 64dot 0deg 1ドットあたり10度
	//4分くらい　240 128ドットなので1ドットあたり2秒
	if (current_millis - last_tempdraw >= 2000) {
		last_tempdraw = current_millis;
		temp_hist[temp_index] = temp;
		temp_index = (temp_index + 1) % 128;
	}
	if (current_millis - last_temptext >= 500) {
		last_temptext = current_millis;
		oled_update(temp);
	}
}

void setrelay(bool turnon) {
	if (turnon && !relay_state) {
		digitalWrite(RELAY_PIN, HIGH);
		relay_state = true;
	} else if (!turnon && relay_state) {
		digitalWrite(RELAY_PIN, LOW);
		relay_state = false;
	}
}

void control_temp(float temp) {
	if (!controlling)
		return;

	if (tdata[control_index].type == TYPE_ON) {
		//指定温度まで継続オン
		if (tdata[control_index].duration > 0 && current_millis - control_start >= tdata[control_index].duration * 1000) {
				control_start = current_millis;
				control_index++;
		}
		else if (temp >= tdata[control_index].on) {
			setrelay(false);
			control_start = current_millis;
			control_index++;
		} else
			setrelay(true);
	}
	else if (tdata[control_index].type == TYPE_OFF) {
		//指定温度までオフ
		if (tdata[control_index].duration > 0 && current_millis - control_start >= tdata[control_index].duration * 1000) {
				control_start = current_millis;
				control_index++;
		}
		else if (temp <= tdata[control_index].on) {
			setrelay(false);
			control_start = current_millis;
			control_index++;
		} else
			setrelay(false);
	}
	else if (tdata[control_index].type == TYPE_KEEP) {
		//指定温度キープ
		if (current_millis - control_start >= tdata[control_index].duration * 1000) {
			control_start = current_millis;
			control_index++;
		}
		else if (temp <= tdata[control_index].on) {
			setrelay(true);
		}
		else if (temp >= tdata[control_index].off) {
			setrelay(false);
		}
	}

	if (control_index >= NUM_DATA) {
		//終了
		controlling = false;
		setrelay(false);
	}
}

void setup() {
	Serial.begin(9600);

	Serial.println("boot");
	pinMode(BUTTON_PIN, INPUT_PULLUP);
	pinMode(RELAY_PIN, OUTPUT);
	digitalWrite(RELAY_PIN, LOW);

	u8g.begin();
	u8g.firstPage();

	Serial.println("start");
}

void loop() {
	current_millis = millis();

	if (current_millis - last_tempcheck >= 100) {
		last_tempcheck = current_millis;

		float temp;
		temp = max(0, gettemp());
		if (temp >= 220){
			if (time_over_220 == 0)
				time_over_220 = millis();
		}else if (time_over_220 != 0)
			time_over_220 = 0;
		
		draw_tempgraph(temp); //per 2 sec
		control_temp(temp);	//per 500ms
	}

	if (buttonpress()) {
		if (!controlling) {
			controlling = true;
			control_index = 0;
			control_start = current_millis;
		} else {
			controlling = false;
			setrelay(false);
		}
	}
}
